
from abc import abstractmethod
import uuid
from datetime import date
from django.db import models
from django.conf import settings
from django.contrib.auth.models import Group


class RequestStatus(models.TextChoices):
    PENDING = 'Pending'
    COMPLETED = 'Completed'


class ApprovalStatus(models.TextChoices):
    APPROVED = 'Approved'
    REJECTED = 'Rejected'


class BaseStage(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, unique=True)
    level = models.IntegerField()
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name
    
    @classmethod
    def first_stage(cls):
        stage = cls.objects.get(level=1)
        if not stage:
            raise Exception("Approval Stage Not Found")
        return stage
    
    @classmethod
    def last_stage(cls):
        stage = cls.objects.order_by("-level").first()
        return stage


class BaseApprover(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, unique=True)
    request_stage = None
    group = models.ForeignKey(Group, related_name="+", on_delete=models.DO_NOTHING)

    class Meta:
        abstract = True

    def __str__(self) -> str:
        return self.group.name
    
    @classmethod
    def get_by_stage(cls, stage):
        if not isinstance(stage, cls):
            stage = cls.objects.get(id=stage)
        return cls.objects.filter(stage=stage)
    
    @classmethod
    def get_by_group(cls, group):
        if not isinstance(group, cls):
            group = cls.objects.get(id=group)
        return cls.objects.filter(group=group)
    

class BaseRequest(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, unique=True)
    request_stage = None
    request_status = models.CharField(max_length=255,choices=RequestStatus.choices, default=RequestStatus.PENDING, blank=True, null=True)
    request_date = models.DateField(auto_now=True, blank=True, null=True)
    request_summary = models.TextField(blank=True, null=True)
    approval_status = models.CharField(max_length=255,choices=ApprovalStatus.choices, default=RequestStatus.PENDING, blank=True, null=True)
    approval_date = models.DateField(blank=True, null=True)

    class Meta:
        abstract = True

    @abstractmethod
    def get_last_stage(self):
        pass
    
    def is_last_stage(self):
        last_stage = self.get_last_stage()
        return self.request_stage == last_stage

    def approve(self):
        self.approval_status = ApprovalStatus.APPROVED
        self.approval_date = date.today()

    def reject(self):
        self.approval_status = ApprovalStatus.REJECTED
        self.approval_date = date.today()
        
    def complete(self):
        self.request_status = RequestStatus.COMPLETED
    
    def is_pending(self):
        return self.request_status==RequestStatus.PENDING

    def is_completed(self):
        return self.request_status==RequestStatus.COMPLETED

    def is_approved(self):
        return self.approval_status==ApprovalStatus.APPROVED

    def is_rejected(self):
        return self.approval_status==ApprovalStatus.REJECTED
   

class BaseApproval(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False, unique=True)
    request = None
    request_stage = None
    decision = models.CharField(max_length=255, choices=ApprovalStatus.choices)
    comment = models.CharField(max_length=255, blank=True, null=True)
    approver = models.OneToOneField(settings.AUTH_USER_MODEL, related_name="+", on_delete=models.CASCADE)
    approval_date = models.DateField(auto_now=True)

    class Meta:
        abstract = True

    def is_approved(self):
        return self.decision==ApprovalStatus.APPROVED

    def is_rejected(self):
        return self.decision==ApprovalStatus.REJECTED





class RequestExistsException(Exception):
    message = "Transfer request exists"

    def __str__(self) -> str:
        return self.message
    

class PermissionDeniedException(Exception):
    message = "Permission denied"

    def __str__(self) -> str:
        return self.message




from abc import ABC, abstractmethod
from django.contrib.auth.models import User
from .exceptions import PermissionDeniedException
from .models import ApprovalStatus, BaseApproval, BaseApprover, BaseRequest, BaseStage, RequestStatus


class ApproverService(ABC):
    approvers: BaseApprover = None
    requests: BaseRequest = None
    approvals: BaseApproval = None
  
    def __init__(self, approver: User) -> None:
        self.approver = approver

    def get_approver(self) -> BaseApprover:
        user_group = self.approver.groups.all()
        return self.approvers.filter(group__in=user_group).first()
    
    def get_requests(self):
        approver = self.get_approver()
        if not approver:
            raise PermissionDeniedException
        
        return self.requests.filter(
            request_status=RequestStatus.PENDING,
            approval_status=RequestStatus.PENDING,
            request_stage=approver.request_stage
        )
    
    def get_history(self):
        approver = self.get_approver()
        if not approver:
            raise PermissionDeniedException
        
        return self.approvals.filter(request_stage=approver.request_stage)

    
class ApprovalService(ABC):
    approvers: BaseApprover = None
    requests: BaseRequest = None
    approvals: BaseApproval = None
    stages: BaseStage = None

    def __init__(self, approver: User, request: BaseRequest, decision: str, comment: str):
        self.approver = approver
        self.request = request
        self.decision = decision
        self.comment = comment
    
    def get_decision(self):
        if str(self.decision).lower() == (ApprovalStatus.APPROVED).lower():
            return ApprovalStatus.APPROVED
        if str(self.decision).lower() == (ApprovalStatus.REJECTED).lower():
            return ApprovalStatus.REJECTED
        
    def create_approval(self):
        approval = self.approvals
        approval.request = self.request
        approval.request_stage = self.request.request_stage
        approval.approver = self.approver
        approval.comment = self.comment
        approval.decision = self.get_decision()
        approval.save()
        return approval

    def get_next_stage(self):
        current_level = self.request.request_stage.level
        return self.stages.filter(level__gt=current_level).first()

    def approve_request(self):
        if self.request.is_last_stage():
            self.request.approve()
            self.request.complete()
            self.request.save()
            self.post_request_approve()
        else:
            next_stage = self.get_next_stage()
            self.request.request_date = next_stage
            self.request.save()

    def reject_request(self):
        request = self.request
        request.reject()
        request.complete()
        request.save()
        self.post_request_reject()

    def process(self):
        approval = self.create_approval()
        if approval.is_approved():
            self.approve_request()
        elif approval.is_rejected():
            self.reject_request()
        return approval
    

    @abstractmethod
    def post_request_approve(self):
       pass
    
    @abstractmethod
    def post_request_reject(self):
       pass


        






from django import forms
from .models import ApprovalStatus


class ApprovalForm(forms.Form):

    comment = forms.CharField(required=True, label="Comment")
    decision = forms.ChoiceField(choices=ApprovalStatus.choices, required=True, label="Decision")

    class Meta:
        fields = ["comment", "decision"]
   
